#include "ble.h"

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)
#define COMPANY_ID_CODE 0x0059

struct bt_conn *ble_connection;
bool hc_ble_lost_con = false;

static uint16_t id_adv[] = {COMPANY_ID_CODE, 0x0000};

// Advertising data definition
static struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA_BYTES(BT_DATA_GAP_APPEARANCE,
                  (CONFIG_BT_DEVICE_APPEARANCE >> 0) & 0xff,
                  (CONFIG_BT_DEVICE_APPEARANCE >> 8) & 0xff),
    BT_DATA(BT_DATA_MANUFACTURER_DATA, (uint8_t *)&id_adv, sizeof(id_adv)),
};

// Scan response definition
static struct bt_data sd[] = {
    BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};

/**
 * @brief Callback for BLE paramaters update event
 * @param conn Pointer to used BLE connection structure
 * @param interval Interval value. See BLE documentation for more information
 * @param latency Latency value. See BLE documentation for more information
 * @param timeout Timeout value. See BLE documentation for more information
 */
static void le_param_updated(struct bt_conn *conn, uint16_t interval,
                             uint16_t latency, uint16_t timeout) {
  struct bt_conn_info info;
  char addr[BT_ADDR_LE_STR_LEN];

  if (bt_conn_get_info(conn, &info)) {
    printk("Could not parse connection info\n");
  } else {
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    printk("Connection parameters updated!	\n\
        Connected to: %s						\n\
        New Connection Interval: %u				\n\
        New Slave Latency: %u					\n\
        New Connection Supervisory Timeout: %u	\n",
           addr, info.le.interval, info.le.latency, info.le.timeout);
  }
}

/**
 * @brief Callback for BLE disconnection event
 * @param conn Pointer to used BLE connection structure
 * @param reason Disconnection reason.  See BLE documentation for more
 * information
 */
static void disconnected(struct bt_conn *conn, uint8_t reason) {
  printk("Disconnected (reason %u)\n", reason);
  bt_conn_unref(conn);
  hc_bt_stop();
}

/**
 * @brief Callback for BLE parameter request event
 * @param conn Pointer to used BLE connection structure
 * @param param Pointer to desired paramter
 * @return True always, this function is not used for the moment
 */
static bool le_param_req(struct bt_conn *conn, struct bt_le_conn_param *param) {
  // If acceptable params, return true, otherwise return false.
  return true;
}

/**
 * @brief Callback fo BLE register event
 *          This function is not used for the moment
 */
static void on_register() { printk("<Inf> BLE register callback called\n"); }

/**
 * @brief Callback for BLE connection event
 * @param conn Pointer to used BLE connection structure
 * @param err Received error code. 0 if no error occured
 */
static void connected(struct bt_conn *conn, uint8_t err) {
  struct bt_conn_info info;
  char addr[BT_ADDR_LE_STR_LEN];

  ble_connection = conn;
  ble_connection = bt_conn_ref(conn);

  if (err) {
    printk("Connection failed (err %u)\n", err);
    return;
  }

  if (bt_conn_get_info(conn, &info)) {
    printk("Could not parse connection info\n");
  } else {
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    printk("Connection established!		\n\
        Connected to: %s					\n\
        Role: %u							\n\
        Connection interval: %u				\n\
        Slave latency: %u					\n\
        Connection supervisory timeout: %u	\n",
           addr, info.role, info.le.interval, info.le.latency, info.le.timeout);
  }
}

int hc_bt_start() {
  int8_t err =
      bt_le_adv_start(BT_LE_ADV_CONN, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));

  if (err) {
    printk("Failed to start advertising (%d)\n", err);
    return err;
  }
  return 0;
}

int hc_bt_stop() {
  // Remove all remaining connections and stop advertising
  bt_conn_foreach(BT_CONN_TYPE_LE, bt_conn_disconnect,
                  (ble_connection, BT_HCI_ERR_REMOTE_USER_TERM_CONN));

  int err = bt_le_adv_stop();
  if (err) {
    printk("Failed to stop advertising (%d)\n", err);
  }
  hc_bt_start();
  return err;
}

static struct bt_conn_cb conn_callbacks = {.connected = connected,
                                           .disconnected = disconnected,
                                           .le_param_req = le_param_req,
                                           .le_param_updated =
                                               le_param_updated};

/**
 * @brief Callback for BLE interface activation
 * @param erro Encountered error code. 0 if no error occured
 */
static void bt_ready(int erro) {
  int err = 0;
  if (erro) {
    printk("BLE init failed with error code %d\n", err);
    return;
  }
  // Configure connection callbacks
  bt_conn_cb_register(&conn_callbacks);
  k_sem_give(&ble_init_ok);
}

int hc_bt_init() {
  int err = 0;
  uint8_t mac_buff[2];
  err = bt_enable(bt_ready);

  if (err) {
    printk("<Err> BLE initialization failed\n");
    return -1;
  }

  err = k_sem_take(&ble_init_ok, K_MSEC(500));

  if (!err) {
    printk("<Inf> Bluetooth initialized\n");
  } else {
    printk("<Err> BLE initialization did not complete in time\n");
  }
  return err;
}